/*
                                                                
Copyright 2006 by BBN Technologies and University of Maryland (UMD)

BBN and UMD grant a nonexclusive, source code, royalty-free right to
use this Software known as Translation Error Rate COMpute (the
"Software") solely for research purposes. Provided, you must agree
to abide by the license and terms stated herein. Title to the
Software and its documentation and all applicable copyrights, trade
secrets, patents and other intellectual rights in it are and remain
with BBN and UMD and shall not be used, revealed, disclosed in
marketing or advertisement or any other activity not explicitly
permitted in writing.

BBN and UMD make no representation about suitability of this
Software for any purposes.  It is provided "AS IS" without express
or implied warranties including (but not limited to) all implied
warranties of merchantability or fitness for a particular purpose.
In no event shall BBN or UMD be liable for any special, indirect or
consequential damages whatsoever resulting from loss of use, data or
profits, whether in an action of contract, negligence or other
tortuous action, arising out of or in connection with the use or
performance of this Software.

Without limitation of the foregoing, user agrees to commit no act
which, directly or indirectly, would violate any U.S. law,
regulation, or treaty, or any other international treaty or
agreement to which the United States adheres or with which the
United States complies, relating to the export or re-export of any
commodities, software, or technical data.  This Software is licensed
to you on the condition that upon completion you will cease to use
the Software and, on request of BBN and UMD, will destroy copies of
the Software in your possession.                                                

TERtest.java v1
Matthew Snover (snover@cs.umd.edu)                           

*/

import java.io.*;

public class TERtest {
    public static void main(String[] args) {
	if (args.length != 2) {
	    System.out.println("usage: java TERtest <hypfile> <reffile>");
	    System.exit(-1);
	}
	double TOTAL_EDITS = 0.0;
	double TOTAL_WORDS = 0.0;

	TERcost costfunc = new TERcost();

	BufferedReader hypstream;
	BufferedReader refstream;

	try {
	    hypstream = 
		new BufferedReader(new FileReader (args[0]));
	    refstream = 
		new BufferedReader(new FileReader (args[1]));
	} catch(IOException ioe) {
	    System.out.println(ioe);
	    return;
	}

	try {
	    String hyp;
	    String ref;
	    while (((hyp = hypstream.readLine()) != null) &&
		   ((ref = refstream.readLine()) != null)) {
		TERalignment result = TERcalc.TER(hyp, ref, costfunc);
		TOTAL_EDITS += result.numEdits;
		TOTAL_WORDS += result.numWords;
		// System.out.println(result.numEdits + "\t" + result.numWords + "\t" + result.score());
		System.out.println("" + result + "\n");
	    }
	} catch(IOException ioe) {
	    System.out.println(ioe);
	    return;
	}
	System.out.println("Total TER: " + (TOTAL_EDITS / TOTAL_WORDS) + " (" +
			   TOTAL_EDITS + "/" + TOTAL_WORDS + ")");
	System.out.println("Number of calls to beam search: " + 
			   TERcalc.numBeamCalls());
	System.out.println("Number of segments scored: " + 
			   TERcalc.numSegsScored());
	System.out.println("Number of shifts tried: " + 
			   TERcalc.numShiftsTried());
    }
}